from flask import Response
from flask import request
import requests


def handle(request_path, details, route):
    url_repl = request.host_url.rstrip('/') + '/' + route['path'].strip('/')
    url = request.url.replace(url_repl, details['host'].rstrip('/'))
    #
    print('proxying to url ' + url)

    resp = requests.request(
        method=request.method,
        url=url,
        headers={key: value for (key, value) in request.headers if key != 'Host'},
        data=request.get_data(),
        cookies=request.cookies,
        allow_redirects=False)

    excluded_headers = ['content-encoding', 'content-length', 'transfer-encoding', 'connection']
    headers = [(name, value) for (name, value) in resp.raw.headers.items()
               if name.lower() not in excluded_headers]

    response = Response(resp.content, resp.status_code, headers)
    return response



import json
from pathvalidate import ValidationError, validate_filename


def fill_config_defaults_static(details):
    if not 'default_file' in details:
        details['default_file'] = 'index.html'
    if not 'as_attachment' in details:
        details['as_attachment'] = False


def fill_main_config_defaults(config):
    for route in config['routes']:        
        details = route['details']
        if route['type'] == 'static':
            fill_config_defaults_static(details)


def fill_cgi_config_defaults(cgi_config):
    if 'cgi' not in cgi_config or len(cgi_config['cgi']) == 0:
        cgi_config['cgi'] = [{ 'extension': 'py', 'cmd': 'python'}]
        return

    for cgi in cgi_config['cgi']:
        if 'extension' not in cgi or 'cmd' not in cgi:
            raise Exception('Error in CGI config')
        validate_filename(cgi_config['cgi']['extension'].lstrip('.'))


def fill_config_defaults(config, cgi_config):
    fill_main_config_defaults(config)
    if len(cgi_config) == 0: 
        cgi_config['cgi'] = []
    fill_cgi_config_defaults(cgi_config)


def read_file(path):
    file = open(path, 'r')
    contents = file.read()
    return json.loads(contents)


def read(path, cgi_path=None):
    config = read_file(path)
    cgi_config = {}
    if not cgi_path is None:
        cgi_config = read_file(cgi_path)
    fill_config_defaults(config, cgi_config)
    return { 'data': config, 'cgi': cgi_config['cgi'] }



import os
from flask import send_file
from flask import abort


def path_replace(target, source):
    target.strip('/')
    source.strip('/')
    target_parts = target.split('/')
    source_parts = source.split('/')
    return '/'.join(target_parts[len(source_parts):]).strip('/')


def get_static_path(request_path, details, route):
    file = None

    clean_path = route['path'].strip('/')

    request_path = path_replace(request_path, clean_path)
    request_path = request_path.strip('/')
    
    # if config details path is file then any request should return this file
    if os.path.isfile(details['path']):
        file = details['path']
    # if config details path is dir then join the request string
    elif os.path.isdir(details['path']):
        file = details['path'].rstrip('/') + '/' + request_path
        # if resulting path is dir then use default file in this dir
        if os.path.isdir(file):
            file = file.rstrip('/') + '/' + details['default_file']
    return file


def handle(request_path, details, route):
    
    file = get_static_path(request_path, details, route)

    if file is None or not os.path.exists(file):
        print('Not found file at ' + file)
        abort(404)
    return send_file(file, as_attachment=details['as_attachment'])



from flask import Response
from ..handlers import static
from werkzeug import Headers
import subprocess


def compose_response(data):
    heading_data_list = []
    body_seek = 0
    for i in range(0, len(data) - 1):
        if chr(data[i]) == '\n' and chr(data[i+1]) == '\n':
            body_seek = i + 2
            break
        heading_data_list.append(chr(data[i]))
    
    heading_data = ''.join(heading_data_list)
    # RFC-2616
    # https://www.w3.org/Protocols/rfc2616/rfc2616-sec6.html#sec6.1
    hd_split = heading_data.split('\n', 1)

    if len(hd_split) < 1:
        raise Exception('bad cgi response')

    head = hd_split[0].strip('\r')

    nohead = ''
    if len(hd_split) > 1:
        nohead = hd_split[1].strip('\n')

    head_values = head.split(' ')
    if len(head_values) < 2:
        raise Exception('bad cgi head: ' + head)
    try:
        status = int(head_values[1]) # status code
    except ValueError:
        raise Exception('bad cgi status code: ' + head_values[1])

    try:
        reason = head_values.split(' ', 2)[2]
    except Exception:
        reason = ''

    headers_split = nohead.split('\n\n', 1)
    headers_str = headers_split[0]

    headers_str_list = headers_str.split('\n')
    headers = Headers()

    disposition_defined = False
    for h_str in headers_str_list:
        if ':' in h_str:
            header_list = h_str.split(':', 1)
            header_key = header_list[0].strip()
            header_value = header_list[1].strip()
            if header_key.lower() == 'content-disposition':
                disposition_defined = True
            headers.add(header_key, header_value)

    if not disposition_defined:
        headers.add('Content-Disposition', 'inline')
        
    response = Response(headers = headers, status = status)
    response.set_data(data[body_seek:].decode("utf-8") )

    return response


def handle(request_path, details, route, cgi_conf):
    file = static.get_static_path(request_path, details, route)

    for c in cgi_conf:
        if file.endswith('.' + c['extension']):
            arglist = c['cmd'].split(' ')
            arglist.append(file)
            print(arglist)
            data = subprocess.run(arglist, capture_output=True).stdout
            return compose_response(data)



import os
import json
from flask import Flask
from flask import request
from flask import abort
from werkzeug.routing import Rule

from .router import resolver
from .configurator import reader as config_reader
from .handlers import static
from .handlers import proxy
from .handlers import text
from .handlers import cgi

name = 'cmdline-mocksrv'


def create_app(test_config=None):

    config_path = os.environ.get('MOCKSRV_CONFIG')
    cgi_config_path = os.environ.get('MOCKSRV_CGI_CONFIG')

    if config_path is None or not os.path.exists(config_path):
        print('MOCKSRV_CONFIG variable is empty')
        print('Please set config in MOCKSRV_CONFIG env variable')
        return

    if cgi_config_path is not None and cgi_config_path != '' and not os.path.exists(cgi_config_path):
        print('MOCKSRV_CGI_CONFIG variable not empty and file doesnt exist')
        print('set MOCKSRV_CGI_CONFIG="" to use default or provide a valid config path')
        return

    m_config = config_reader.read(config_path, cgi_config_path)
    app = Flask(__name__, static_folder=None, template_folder=None)

    # using werkzeug routing for cathing any http method
    app.url_map.add(Rule('/', endpoint='catch_query'))
    app.url_map.add(Rule('/<path:path>', endpoint='catch_query'))

    @app.endpoint('catch_query')
    def catch_query(path=''):
        route = resolver.resolve(path, m_config['data'])

        if route is None:
            abort(404)

        print('Resolved route: ' + route.get('path'))

        route_type = route['type']

        if route_type == 'static':
            return static.handle(path, route['details'], route)
        elif route_type == 'proxy':
            return proxy.handle(path, route['details'], route)
        elif route_type == 'text':
            return text.handle(path, route['details'], route)
        elif route_type == 'cgi':
            return cgi.handle(path, route['details'], route, m_config['cgi'])

    return app



from flask import Response
from flask import request
import json

def get_text_response(content_obj):
    json_data = json.dumps(content_obj)
    str = json_data.strip('"')
    mimetype = 'application/json'
    if len(str) < len(json_data):
        mimetype = 'text/plain'
    return Response(str, mimetype=mimetype)


def handle(request_path, details, route):
    method = str(request.method)

    if 'var' in details:
        for rule in details['var']:
            if '_verb' in rule and rule['_verb'] != method:
                continue
            matched = True
            for name in rule.keys():
                if name == '_verb' or name == '_content':
                    continue
                else:
                    value = rule[name]
                    comps = route['path'].strip('/').split('/')
                    p_comps = request_path.strip('/').split('/')
                    current_value = None
                    for i in range(0, len(comps)):
                        if comps[i].startswith('{' + name + "}") or comps[i].startswith('{' + name + ':'):
                            current_value = p_comps[i]
                    if value != current_value:
                        matched = False
                        continue

            if matched:
                return get_text_response(rule['_content'])

    if not 'verb' in details:
         return get_text_response(details['content'])

    if not method in details['verb'].keys():
        if not '*' in details['verb'].keys():
             return get_text_response(details['content'])
        return get_text_response(details['verb']['*'])
    return get_text_response(details['verb'][method])


def is_ascii(s):
    return all(ord(c) < 128 for c in s)

def split_path(path):
    return list(filter(None, path.split('/')))


def check_braces(path):
    braces = []
    for symb in path:
        if symb == '{':
            braces.append('{')
        elif symb == '}':
            if len(braces) == 0:
                raise Exception('invalid route in config')
            braces.pop()
    if len(braces) != 0:
        raise Exception('invalid route in config')


def pattern_match(pattern, value):
    if not pattern.startswith('{') or not pattern.endswith('}') or pattern.count(':') > 1 or pattern.count('/') > 0:
        raise Exception('Invalid pattern. pattern: ' + pattern)

    parts = pattern.strip('{}').split(':')
    if len(parts) == 1:
        return True

    restriction = parts[1].lower()
    if restriction == 'number':
        return value.lstrip('-').isdigit()
    elif restriction == 'text':
        return True
    elif restriction == 'ascii':
        return is_ascii(value)


def resolve(path, config, depth=0):
    if depth > 5:
        raise Exception('Too many map redirects in config')
    best_route = None
    path_parts = split_path(path)

    best_len = -1
    for route in config['routes']:
        route_parts = split_path(route['path'])
        if len(route_parts) > len(path_parts):
            continue
        common_len = 0
        match = True
        for i in range(0, len(route_parts)):
            if route_parts[i] == path_parts[i] or route_parts[i].startswith('{'):
                if route_parts[i].startswith('{') and not pattern_match(route_parts[i], path_parts[i]):
                    match = False
                common_len = common_len + 1
            else:
                match = False

        if match and common_len > best_len:
            best_len = common_len
            best_route = route

    if best_route is not None and best_route['type'] == 'map':
        best_parts = split_path(best_route['path'])
        new_path = '/'.join(split_path(best_route['details']['to']) + path_parts[best_len:])
        return resolve(new_path, config, depth+1)
    return best_route


def validate_config(config):
    for route in config['routes']:
        check_braces(route['path'])



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='cmdline-mocksrv',
    version='1.0.0',
    author="Nikita Tarasov",
    author_email="mail@ntarasov.ru",
    description="Mocking service",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nntarasov/mocksrv",
    packages={'mocksrv', 'mocksrv.configurator', 'mocksrv.handlers', 'mocksrv.router'},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'flask', 'pathvalidate'
    ],
)
